@@ -29,7 +29,7 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
- * $Id: exif.c,v 1.1.1.1 2003/08/07 16:46:05 ccpro Exp $
+ * $Id: exif.c,v 1.60 2003/08/16 03:34:21 ejohnst Exp $
  */
 
 /*
@@ -57,13 +57,13 @@
 #include "makers.h"
 
 #define OLYMPUS_BUGS		/* Work around Olympus stupidity. */
-#define FUJI_BUGS		/* Work around Fuji stupidity. */
 #define WINXP_BUGS		/* Work around Windows XP stupidity. */
+#define UNCREDITED_BUGS		/* Work around uncredited stupidity. */
 
 
 /* Function prototypes. */
 
-static int parsetag(struct exifprop *prop, struct ifd *dir,
+static void parsetag(struct exifprop *prop, struct ifd *dir,
     struct exiftags *t, int domkr);
 
 
@@ -76,42 +76,56 @@ readtag(struct field *afield, int ifdseq, struct ifd *dir, struct exiftags *t,
 {
 	int i, j;
 	struct exifprop *prop, *tmpprop;
+	u_int16_t tag;
 
 	prop = newprop();
+	if (dir->par)
+		tag = dir->par->tag;
+	else
+		tag = EXIF_T_UNKNOWN;
 
 	/* Field info. */
 
-	prop->tag = exif2byte(afield->tag, t->tifforder);
-	prop->type = exif2byte(afield->type, t->tifforder);
-	prop->count = exif4byte(afield->count, t->tifforder);
+	prop->tag = exif2byte(afield->tag, dir->md.order);
+	prop->type = exif2byte(afield->type, dir->md.order);
+	prop->count = exif4byte(afield->count, dir->md.order);
 	if ((prop->type == TIFF_SHORT || prop->type == TIFF_SSHORT) &&
 	    prop->count <= 1)
-		prop->value = exif2byte(afield->value, t->tifforder);
+		prop->value = exif2byte(afield->value, dir->md.order);
 	else
-		prop->value = exif4byte(afield->value, t->tifforder);
+		prop->value = exif4byte(afield->value, dir->md.order);
 
 	/* IFD identifying info. */
 
 	prop->ifdseq = ifdseq;
-	prop->ifdtag = dir->tag;
+	prop->par = dir->par;
+	prop->tagset = dir->tagset;
 
 	/* Lookup the field name. */
 
-	for (i = 0; tags[i].tag < EXIF_T_UNKNOWN &&
-	    tags[i].tag != prop->tag; i++);
-	prop->name = tags[i].name;
-	prop->descr = tags[i].descr;
-	prop->lvl = tags[i].lvl;
+	for (i = 0; prop->tagset[i].tag < EXIF_T_UNKNOWN &&
+	    prop->tagset[i].tag != prop->tag; i++);
+	prop->name = prop->tagset[i].name;
+	prop->descr = prop->tagset[i].descr;
+	prop->lvl = prop->tagset[i].lvl;
 
-	/* Lookup and check the field type. */
+	/*
+	 * Lookup and check the field type.
+	 *
+	 * We have to be pretty severe with entries that have an invalid
+	 * field type -- too many assumptions in the rest of the code.
+	 */
 
 	for (j = 0; ftypes[j].type && ftypes[j].type != prop->type; j++);
-	if (!ftypes[j].type)
-		exifdie("unknown TIFF field type");
+	if (!ftypes[j].type) {
+		exifwarn2("unknown TIFF field type; discarding", prop->name);
+		free(prop);
+		return;
+	}
 
 	/* Skip sanity checking on maker note tags; we'll get to them later. */
 
-	if (prop->ifdtag != EXIF_T_MAKERNOTE) {
+	if (tag != EXIF_T_MAKERNOTE) {
 		/*
 		 * XXX Ignore UserComment -- a hack to get around an apparent
 		 * WinXP Picture Viewer bug (err, liberty).  When you rotate
@@ -120,7 +134,7 @@ readtag(struct field *afield, int ifdseq, struct ifd *dir, struct exiftags *t,
 		 * (At least we're able to ID invalid comments...)
 		 */
 
-		if (tags[i].type && tags[i].type != prop->type
+		if (prop->tagset[i].type && prop->tagset[i].type != prop->type
 #ifdef WINXP_BUGS
 		    && prop->tag != EXIF_T_USERCOMMENT
 #endif
@@ -129,7 +143,8 @@ readtag(struct field *afield, int ifdseq, struct ifd *dir, struct exiftags *t,
 
 		/* Check the field count. */
 
-		if (tags[i].count && tags[i].count != prop->count)
+		if (prop->tagset[i].count && prop->tagset[i].count !=
+		    prop->count)
 			exifwarn2("field count mismatch", prop->name);
 	}
 
@@ -139,18 +154,16 @@ readtag(struct field *afield, int ifdseq, struct ifd *dir, struct exiftags *t,
 
 	/*
 	 * Do as much as we can with the tag at this point and add it
-	 * to our list if it's not an IFD pointer.
+	 * to our list.
 	 */
 
-	if (parsetag(prop, dir, t, domkr)) {
-		if ((tmpprop = t->props)) {
-			while (tmpprop->next)
-				tmpprop = tmpprop->next;
-			tmpprop->next = prop;
-		} else
-			t->props = prop;
+	parsetag(prop, dir, t, domkr);
+	if ((tmpprop = t->props)) {
+		while (tmpprop->next)
+			tmpprop = tmpprop->next;
+		tmpprop->next = prop;
 	} else
-		free(prop);
+		t->props = prop;
 }
 
 
@@ -162,28 +175,16 @@ readtag(struct field *afield, int ifdseq, struct ifd *dir, struct exiftags *t,
  * the count here.  Root IFDs (0 and 1) are processed first (along with any
  * other "root" IFDs we find), then any nested IFDs in the order they're
  * encountered.
- *
- * XXX To work around some Fuji braindeadness, if we're dealing with the
- * maker note we set our byte order to little endian.
  */
 static void
 readtags(struct ifd *dir, int seq, struct exiftags *t, int domkr)
 {
 	int i;
-#ifdef FUJI_BUGS
-	enum order tmporder;
-
-	tmporder = t->tifforder;
-	if (dir->tag == EXIF_T_MAKERNOTE && t->mkrval == EXIF_MKR_FUJI)
-		t->tifforder = LITTLE;
-#endif
 
 	if (debug) {
-		if (dir->tag != EXIF_T_UNKNOWN) {
-			for (i = 0; tags[i].tag < EXIF_T_UNKNOWN &&
-			    tags[i].tag != dir->tag; i++);
+		if (dir->par && dir->par->tag != EXIF_T_UNKNOWN) {
 			printf("Processing %s directory, %d entries\n",
-			    tags[i].name, dir->num);
+			    dir->par->name, dir->num);
 		} else
 			printf("Processing directory %d, %d entries\n",
 			    seq, dir->num);
@@ -192,10 +193,6 @@ readtags(struct ifd *dir, int seq, struct exiftags *t, int domkr)
 	for (i = 0; i < dir->num; i++)
 		readtag(&(dir->fields[i]), seq, dir, t, domkr);
 
-#ifdef FUJI_BUGS
-	if (dir->tag == EXIF_T_MAKERNOTE && t->mkrval == EXIF_MKR_FUJI)
-		t->tifforder = tmporder;
-#endif
 	if (debug)
 		printf("\n");
 }
@@ -205,6 +202,9 @@ readtags(struct ifd *dir, int seq, struct exiftags *t, int domkr)
  * Post-process property values.  By now we've got all of the standard
  * Exif tags read in (but not maker tags), so it's safe to work out
  * dependencies between tags.
+ *
+ * XXX At this point, we've lost IFD-level TIFF metadata.  Therefore,
+ * assumptions about byte order and beginning of the TIFF might be false.
  */
 static void
 postprop(struct exifprop *prop, struct exiftags *t)
@@ -213,16 +213,35 @@ postprop(struct exifprop *prop, struct exiftags *t)
 	u_int16_t v;
 	u_int32_t val;
 	float fval;
-	enum order o = t->tifforder;
+	enum byteorder o = t->md.order;
 	struct exifprop *h = t->props;
 
-	/* Process maker note tags specially... */
+	/*
+	 * Process tags from special IFDs.
+	 * XXX Don't properly pass IFD byte order here.
+	 */
 
-	if (prop->ifdtag == EXIF_T_MAKERNOTE && makers[t->mkrval].propfun) {
-		makers[t->mkrval].propfun(prop, t);
-		return;
+	if (prop->par && prop->par->tagset == tags) {
+		switch (prop->par->tag) {
+
+		case EXIF_T_MAKERNOTE:
+			if (makers[t->mkrval].propfun) {
+				makers[t->mkrval].propfun(prop, t);
+				return;
+			}
+			break;
+
+		case EXIF_T_GPSIFD:
+			gpsprop(prop, t);
+			return;
+		}
 	}
 
+	/* Process normal tags. */
+
+	if (prop->tagset != tags)
+		return;
+
 	switch (prop->tag) {
 
 	case EXIF_T_XRES:
@@ -230,12 +249,14 @@ postprop(struct exifprop *prop, struct exiftags *t)
 	case EXIF_T_FPXRES:
 	case EXIF_T_FPYRES:
 		if (prop->tag == EXIF_T_XRES || prop->tag == EXIF_T_YRES) {
-			if (!(tmpprop = findprop(h, EXIF_T_RESUNITS))) break;
+			if (!(tmpprop = findprop(h, tags, EXIF_T_RESUNITS)))
+				break;
 		} else {
-			if (!(tmpprop = findprop(h, EXIF_T_FPRESUNITS))) break;
+			if (!(tmpprop = findprop(h, tags, EXIF_T_FPRESUNITS)))
+				break;
 		}
-		val = exif4byte(t->btiff + prop->value, o) /
-		    exif4byte(t->btiff + prop->value + 4, o);
+		val = exif4byte(t->md.btiff + prop->value, o) /
+		    exif4byte(t->md.btiff + prop->value + 4, o);
 		snprintf(prop->str, 31, "%d dp%s", val, tmpprop->str);
 		prop->str[31] = '\0';
 		break;
@@ -247,8 +268,8 @@ postprop(struct exifprop *prop, struct exiftags *t)
 	 */
 
 	case EXIF_T_SHUTTER:
-		fval = (float)exif4sbyte(t->btiff + prop->value, o) /
-		    (float)exif4sbyte(t->btiff + prop->value + 4, o);
+		fval = (float)exif4sbyte(t->md.btiff + prop->value, o) /
+		    (float)exif4sbyte(t->md.btiff + prop->value + 4, o);
 		if (isnan(fval)) fval = 0;
 		/* 1 / (2^speed) */
 		snprintf(prop->str, 31, "1/%d",
@@ -264,8 +285,8 @@ postprop(struct exifprop *prop, struct exiftags *t)
 		break;
 
 	case EXIF_T_FNUMBER:
-		fval = (float)exif4byte(t->btiff + prop->value, o) /
-		    (float)exif4byte(t->btiff + prop->value + 4, o);
+		fval = (float)exif4byte(t->md.btiff + prop->value, o) /
+		    (float)exif4byte(t->md.btiff + prop->value + 4, o);
 		if (isnan(fval)) fval = 0;
 		snprintf(prop->str, 31, "f/%.1f", fval);
 		prop->str[31] = '\0';
@@ -273,8 +294,8 @@ postprop(struct exifprop *prop, struct exiftags *t)
 
 	case EXIF_T_LAPERTURE:
 	case EXIF_T_MAXAPERTURE:
-		fval = (float)exif4byte(t->btiff + prop->value, o) /
-		    (float)exif4byte(t->btiff + prop->value + 4, o);
+		fval = (float)exif4byte(t->md.btiff + prop->value, o) /
+		    (float)exif4byte(t->md.btiff + prop->value + 4, o);
 		if (isnan(fval)) fval = 0;
 		/* sqrt(2)^aperture */
 		snprintf(prop->str, 31, "f/%.1f", pow(1.4142, (double)fval));
@@ -282,7 +303,7 @@ postprop(struct exifprop *prop, struct exiftags *t)
 		break;
 
 	case EXIF_T_BRIGHTVAL:
-		if (exif4byte(t->btiff + prop->value, o) == 0xffffffff) {
+		if (exif4byte(t->md.btiff + prop->value, o) == 0xffffffff) {
 			strcpy(prop->str, "Unknown");
 			break;
 		}
@@ -294,16 +315,16 @@ postprop(struct exifprop *prop, struct exiftags *t)
 		break;
 
 	case EXIF_T_DISTANCE:
-		if (exif4byte(t->btiff + prop->value, o) == 0xffffffff) {
+		if (exif4byte(t->md.btiff + prop->value, o) == 0xffffffff) {
 			strcpy(prop->str, "Infinity");
 			break;
 		}
-		if (exif4byte(t->btiff + prop->value + 4, o) == 0) {
+		if (exif4byte(t->md.btiff + prop->value + 4, o) == 0) {
 			strcpy(prop->str, "Unknown");
 			break;
 		}
-		fval = (float)exif4byte(t->btiff + prop->value, o) /
-		    (float)exif4byte(t->btiff + prop->value + 4, o);
+		fval = (float)exif4byte(t->md.btiff + prop->value, o) /
+		    (float)exif4byte(t->md.btiff + prop->value + 4, o);
 		if (isnan(fval)) fval = 0;
 		snprintf(prop->str, 31, "%.2f m", fval);
 		prop->str[31] = '\0';
@@ -320,8 +341,8 @@ postprop(struct exifprop *prop, struct exiftags *t)
 		break;
 
 	case EXIF_T_FOCALLEN:
-		fval = (float)exif4byte(t->btiff + prop->value, o) /
-		    (float)exif4byte(t->btiff + prop->value + 4, o);
+		fval = (float)exif4byte(t->md.btiff + prop->value, o) /
+		    (float)exif4byte(t->md.btiff + prop->value + 4, o);
 		if (isnan(fval)) fval = 0;
 		snprintf(prop->str, 31, "%.2f mm", fval);
 		prop->str[31] = '\0';
@@ -330,19 +351,27 @@ postprop(struct exifprop *prop, struct exiftags *t)
 	/* Digital zoom: set to verbose if numerator is 0 or fraction = 1. */
 
 	case EXIF_T_DIGIZOOM:
-		if (!exif4byte(t->btiff + prop->value, o))
+		if (!exif4byte(t->md.btiff + prop->value, o))
 			strcpy(prop->str, "Unused");
-		else if (exif4byte(t->btiff + prop->value, o) !=
-		    exif4byte(t->btiff + prop->value + 4, o))
+		else if (exif4byte(t->md.btiff + prop->value, o) !=
+		    exif4byte(t->md.btiff + prop->value + 4, o))
 			break;
 		prop->lvl = ED_VRB;
 		break;
 
 	case EXIF_T_FOCALLEN35:
-		if (!(prop->str = (char *)malloc(16)))
-			exifdie((const char *)strerror(errno));
+		exifstralloc(&prop->str, 16);
 		snprintf(prop->str, 15, "%d mm", prop->value);
-		prop->str[15] = '\0';
+		break;
+
+	/*
+	 * XXX This really should be in parsetag() to guarantee that it's
+	 * done before we process the maker notes.  However, I haven't seen
+	 * model not come first, so it should be safe (and more convenient).
+	 */
+
+	case EXIF_T_MODEL:
+		t->model = prop->str;
 		break;
 	}
 }
@@ -373,7 +402,10 @@ tweaklvl(struct exifprop *prop, struct exiftags *t)
 	if (prop->ifdseq == 1 && prop->lvl != ED_UNK)
 		prop->lvl = ED_VRB;
 
-	if (prop->override && (tmpprop = findprop(t->props, prop->override)))
+	/* Maker tags can override normal Exif tags. */
+
+	if (prop->override && (tmpprop = findprop(t->props, tags,
+	    prop->override)))
 		if (tmpprop->lvl & (ED_CAM | ED_IMG | ED_PAS))
 			tmpprop->lvl = ED_OVR;
 }
@@ -382,7 +414,7 @@ tweaklvl(struct exifprop *prop, struct exiftags *t)
 /*
  * Fetch the data for an Exif tag.
  */
-static int
+static void
 parsetag(struct exifprop *prop, struct ifd *dir, struct exiftags *t, int domkr)
 {
 	unsigned int i, len;
@@ -390,16 +422,21 @@ parsetag(struct exifprop *prop, struct ifd *dir, struct exiftags *t, int domkr)
 	u_int32_t un, ud, denom;
 	int32_t sn, sd;
 	char buf[32], *c, *d;
+	struct tiffmeta *md;
+	unsigned char *btiff = dir->md.btiff;
+	enum byteorder o = dir->md.order;
 
 	/* Set description if we have a lookup table. */
 
-	for (i = 0; tags[i].tag < EXIF_T_UNKNOWN &&
-	    tags[i].tag != prop->tag; i++);
-	if (tags[i].table) {
-		prop->str = finddescr(tags[i].table, v);
-		return (TRUE);
+	for (i = 0; prop->tagset[i].tag < EXIF_T_UNKNOWN &&
+	    prop->tagset[i].tag != prop->tag; i++);
+	if (prop->tagset[i].table) {
+		prop->str = finddescr(prop->tagset[i].table, v);
+		return;
 	}
 
+	/* XXX Probably shouldn't process this switch for non-standard tags. */
+
 	switch (prop->tag) {
 
 	/* Process an Exif IFD. */
@@ -407,6 +444,7 @@ parsetag(struct exifprop *prop, struct ifd *dir, struct exiftags *t, int domkr)
 	case EXIF_T_EXIFIFD:
 	case EXIF_T_GPSIFD:
 	case EXIF_T_INTEROP:
+		md = &dir->md;
 		while (dir->next)
 			dir = dir->next;
 
@@ -417,10 +455,14 @@ parsetag(struct exifprop *prop, struct ifd *dir, struct exiftags *t, int domkr)
 		 */
 #ifdef OLYMPUS_BUGS
 		if (prop->tag == EXIF_T_EXIFIFD)
-			readifd(t->btiff + prop->value, &dir->next, t);
+			readifd(prop->value, &dir->next, tags, md);
 		else
 #endif
-			dir->next = readifds(prop->value, t);
+			if (prop->tag == EXIF_T_GPSIFD) {
+				dir->next = readifds(prop->value, gpstags, md);
+			} else {
+				dir->next = readifds(prop->value, tags, md);
+			}
 
 		if (!dir->next) {
 
@@ -438,15 +480,15 @@ parsetag(struct exifprop *prop, struct ifd *dir, struct exiftags *t, int domkr)
 			break;
 		}
 
-		dir->next->tag = prop->tag;
-		return (FALSE);		/* No need to add to property list. */
+		/* XXX Doesn't catch multiple IFDs. */
+		dir->next->par = prop;
+		return;
 
 	/* Record the Exif version. */
 
 	case EXIF_T_VERSION:
 		/* These contortions are to make 0220 = 2.20. */
-		if (!(prop->str = (char *)malloc(8)))
-			exifdie((const char *)strerror(errno));
+		exifstralloc(&prop->str, 8);
 
 		/* Platform byte order affects this... */
 
@@ -456,7 +498,7 @@ parsetag(struct exifprop *prop, struct ifd *dir, struct exiftags *t, int domkr)
 		else
 			for (i = 0; i < 4; i++)
 				buf[i] = ((const char *)&prop->value)[3 - i];
-		buf[4] = prop->str[7] = '\0';
+		buf[4] = '\0';
 		t->exifmin = (short)atoi(buf + 2);
 		buf[2] = '\0';
 		t->exifmaj = (short)atoi(buf);
@@ -467,8 +509,9 @@ parsetag(struct exifprop *prop, struct ifd *dir, struct exiftags *t, int domkr)
 
 	case EXIF_T_MAKERNOTE:
 		if (!domkr)
-			return (TRUE);
+			return;
 
+		md = &dir->md;
 		while (dir->next)
 			dir = dir->next;
 
@@ -481,20 +524,21 @@ parsetag(struct exifprop *prop, struct ifd *dir, struct exiftags *t, int domkr)
 		 */
 
 		if (makers[t->mkrval].ifdfun)
-			dir->next = makers[t->mkrval].ifdfun(prop->value, t);
+			dir->next = makers[t->mkrval].ifdfun(prop->value, md);
 		else
 			exifwarn("maker note not supported");
 
 		if (!dir->next)
 			break;
 
-		dir->next->tag = prop->tag;
-		return (FALSE);		/* No need to add to property list. */
+		/* XXX Doesn't catch multiple IFDs. */
+		dir->next->par = prop;
+		return;
 
 	/* Lookup functions for maker note. */
 
 	case EXIF_T_EQUIPMAKE:
-		strncpy(buf, t->btiff + prop->value, sizeof(buf));
+		strncpy(buf, (const char *)(btiff + prop->value), sizeof(buf));
 		buf[sizeof(buf) - 1] = '\0';
 		for (c = buf; *c; c++) *c = tolower(*c);
 
@@ -511,9 +555,24 @@ parsetag(struct exifprop *prop, struct ifd *dir, struct exiftags *t, int domkr)
 	 * Handle user comment.  According to the spec, the first 8 bytes
 	 * of the comment indicate what charset follows.  For now, we
 	 * just support ASCII.
+	 *
+	 * XXX A handful of the GPS tags are also stored in this format
+	 * (GPSProcessingMethod & GPSAreaInformation).
 	 */
 
+	case 0x001b:
+	case 0x001c:
+		/*
+		 * XXX Note that this is kind of dangerous -- any other
+		 * tag set won't reach the end of the switch...
+		 */
+		if (prop->tagset != gpstags)
+			break;
+		/* FALLTHROUGH */
+
 	case EXIF_T_USERCOMMENT:
+		if (prop->count < 8)
+			break;
 
 		/* Ignore the 'comments' WinXP creates when rotating. */
 #ifdef WINXP_BUGS
@@ -525,27 +584,24 @@ parsetag(struct exifprop *prop, struct ifd *dir, struct exiftags *t, int domkr)
 		/* Lookup the comment type. */
 
 		for (i = 0; ucomment[i].descr; i++)
-			if (!memcmp(ucomment[i].descr,
-			    t->btiff + prop->value, 8))
+			if (!memcmp(ucomment[i].descr, btiff + prop->value, 8))
 				break;
 
 		/* Handle an ASCII comment; strip any trailing whitespace. */
 
 		if (ucomment[i].val == TIFF_ASCII &&
 		    (prop->value + prop->count <
-		    (u_int32_t)(t->etiff - t->btiff))) {
-			c = t->btiff + prop->value + 8;
+		    (u_int32_t)(dir->md.etiff - btiff))) {
+			c = (char *)(btiff + prop->value + 8);
 			d = strlen(c) < prop->count - 8 ? c + strlen(c) :
 			    c + prop->count - 8;
 
 			while (d > c && isspace((int)*(d - 1))) --d;
 
-			if (!(prop->str = (char *)malloc(d - c + 1)))
-				exifdie((const char *)strerror(errno));
+			exifstralloc(&prop->str, d - c + 1);
 			strncpy(prop->str, c, d - c);
-			prop->str[d - c] = '\0';
 			prop->lvl = prop->str[0] ? ED_IMG : ED_VRB;
-			return (TRUE);
+			return;
 		}
 		break;
 
@@ -559,7 +615,7 @@ parsetag(struct exifprop *prop, struct ifd *dir, struct exiftags *t, int domkr)
 #else
 		prop->str = finddescr(filesrcs, v);
 #endif
-		return (TRUE);
+		return;
 	}
 
 	/*
@@ -568,12 +624,12 @@ parsetag(struct exifprop *prop, struct ifd *dir, struct exiftags *t, int domkr)
 	 */
 
 	if (prop->type == TIFF_ASCII &&
-	    (prop->value + prop->count < (u_int32_t)(t->etiff - t->btiff))) {
-		if (!(prop->str = (char *)malloc(prop->count + 1)))
-			exifdie((const char *)strerror(errno));
-		strncpy(prop->str, t->btiff + prop->value, prop->count);
-		prop->str[prop->count] = '\0';
-		return (TRUE);
+	    (prop->value + prop->count <=
+	    (u_int32_t)(dir->md.etiff - btiff))) {
+		exifstralloc(&prop->str, prop->count + 1);
+		strncpy(prop->str, (const char *)(btiff + prop->value),
+		    prop->count);
+		return;
 	}
 
 	/*
@@ -583,25 +639,22 @@ parsetag(struct exifprop *prop, struct ifd *dir, struct exiftags *t, int domkr)
 
 	if ((prop->type == TIFF_RTNL || prop->type == TIFF_SRTNL) &&
 	    (prop->value + prop->count * 8 <=
-	    (u_int32_t)(t->etiff - t->btiff))) {
+	    (u_int32_t)(dir->md.etiff - btiff))) {
 
-		if (!(prop->str = (char *)malloc(32)))
-			exifdie((const char *)strerror(errno));
+		exifstralloc(&prop->str, 32);
 
 		if (prop->type == TIFF_RTNL) {
-			un = exif4byte(t->btiff + prop->value, t->tifforder);
-			ud = exif4byte(t->btiff + prop->value + 4,
-			    t->tifforder);
+			un = exif4byte(btiff + prop->value, o);
+			ud = exif4byte(btiff + prop->value + 4, o);
 			denom = gcd(un, ud);
 			fixfract(prop->str, un, ud, denom);
 		} else {
-			sn = exif4sbyte(t->btiff + prop->value, t->tifforder);
-			sd = exif4sbyte(t->btiff + prop->value + 4,
-			    t->tifforder);
+			sn = exif4sbyte(btiff + prop->value, o);
+			sd = exif4sbyte(btiff + prop->value + 4, o);
 			denom = gcd(abs(sn), abs(sd));
 			fixfract(prop->str, sn, sd, (int32_t)denom);
 		}
-		return (TRUE);
+		return;
 	}
 
 	/*
@@ -613,32 +666,29 @@ parsetag(struct exifprop *prop, struct ifd *dir, struct exiftags *t, int domkr)
 
 	if ((prop->type == TIFF_SHORT || prop->type == TIFF_SSHORT) &&
 	    prop->count > 1 && (prop->value + prop->count * 2 <=
-	    (u_int32_t)(t->etiff - t->btiff))) {
+	    (u_int32_t)(dir->md.etiff - btiff))) {
 
 		if (prop->count > 8)
-			return (TRUE);
-
+			return;
 		len = 8 * prop->count + 1;
-		if (!(prop->str = (char *)malloc(len)))
-			exifdie((const char *)strerror(errno));
-		prop->str[0] = '\0';
+		exifstralloc(&prop->str, len);
 
 		for (i = 0; i < prop->count; i++) {
 			if (prop->type == TIFF_SHORT)
 				snprintf(prop->str + strlen(prop->str),
 				    len - strlen(prop->str) - 1, "%d, ",
-				    exif2byte(t->btiff + prop->value +
-				    (i * 2), t->tifforder));
+				    exif2byte(btiff + prop->value +
+				    (i * 2), o));
 			else
 				snprintf(prop->str + strlen(prop->str),
 				    len - strlen(prop->str) - 1, "%d, ",
-				    exif2sbyte(t->btiff + prop->value +
-				    (i * 2), t->tifforder));
+				    exif2sbyte(btiff + prop->value +
+				    (i * 2), o));
 		}
 		prop->str[strlen(prop->str) - 2] = '\0';
-		return (TRUE);
+		return;
 	}
-	return (TRUE);
+	return;
 }
 
 
@@ -675,12 +725,15 @@ exifscan(unsigned char *b, int len, int domkr)
 	/* Create and initialize our file info structure. */
 
 	t = (struct exiftags *)malloc(sizeof(struct exiftags));
-	if (!t)
-		exifdie((const char *)strerror(errno));
+	if (!t) {
+		exifwarn2("can't allocate file info",
+		    (const char *)strerror(errno));
+		return (NULL);
+	}
 	memset(t, 0, sizeof(struct exiftags));
 
 	seq = 0;
-	t->etiff = b + len;	/* End of TIFF. */
+	t->md.etiff = b + len;	/* End of TIFF. */
 
 	/*
 	 * Make sure we've got the proper Exif header.  If not, we're
@@ -696,21 +749,21 @@ exifscan(unsigned char *b, int len, int domkr)
 	/* Determine endianness of the TIFF data. */
 
 	if (*((u_int16_t *)b) == 0x4d4d)
-		t->tifforder = BIG;
+		t->md.order = BIG;
 	else if (*((u_int16_t *)b) == 0x4949)
-		t->tifforder = LITTLE;
+		t->md.order = LITTLE;
 	else {
 		exifwarn("invalid TIFF header");
 		exiffree(t);
 		return (NULL);
 	}
 
-	t->btiff = b;		/* Beginning of TIFF. */
+	t->md.btiff = b;	/* Beginning of TIFF. */
 	b += 2;
 
 	/* Verify the TIFF header. */
 
-	if (exif2byte(b, t->tifforder) != 42) {
+	if (exif2byte(b, t->md.order) != 42) {
 		exifwarn("invalid TIFF header");
 		exiffree(t);
 		return (NULL);
@@ -719,8 +772,8 @@ exifscan(unsigned char *b, int len, int domkr)
 
 	/* Get the 0th IFD, where all of the good stuff should start. */
 
-	ifdoff = exif4byte(b, t->tifforder);
-	curifd = readifds(ifdoff, t);
+	ifdoff = exif4byte(b, t->md.order);
+	curifd = readifds(ifdoff, tags, &t->md);
 	if (!curifd) {
 		exifwarn("invalid Exif format (couldn't read IFD0)");
 		exiffree(t);
